from typing import Annotated, Literal

import numpy as np
import numpy.typing as npt

from . import Message
from .std_msgs import Header

Float64Array6x6 = Annotated[npt.NDArray[np.float64], Literal[4]]


class Vector3(Message):
    def __init__(self, x: float = 0.0, y: float = 0.0, z: float = 0.0):
        self.x = x  # type: float
        self.y = y  # type: float
        self.z = z  # type: float

    def __str__(self):
        return "Vector3(x={}, y={}, z={})".format(self.x, self.y, self.z)

    def to_json(self):
        return {"x": self.x, "y": self.y, "z": self.z}

    def from_json(self, json):
        self.x = json["x"]
        self.y = json["y"]
        self.z = json["z"]


class Vector3Stamped(Message):
    def __init__(self, header=Header(), vector=Vector3()):
        self.header = header  # type: Header
        self.vector = vector  # type: Vector3

    def __str__(self):
        return "Vector3Stamped(header={}, vector={})".format(self.header, self.vector)

    def to_json(self):
        return {"header": self.header.to_json(), "vector": self.vector.to_json()}

    def from_json(self, json):
        self.header.from_json(json["header"])
        self.vector.from_json(json["vector"])


class Twist(Message):
    def __init__(self, linear=Vector3(), angular=Vector3()):
        self.linear = linear  # type: Vector3
        self.angular = angular  # type: Vector3

    def __str__(self):
        return "Twist(linear={}, angular={})".format(self.linear, self.angular)

    def to_json(self):
        return {"linear": self.linear.to_json(), "angular": self.angular.to_json()}

    def from_json(self, json):
        self.linear.from_json(json["linear"])
        self.angular.from_json(json["angular"])


class Quaternion(Message):
    def __init__(self, x: float = 0.0, y: float = 0.0, z: float = 0.0, w: float = 0.0):
        self.x = x  # type: float
        self.y = y  # type: float
        self.z = z  # type: float
        self.w = w  # type: float

    def __str__(self):
        return "Quaternion(x={}, y={}, z={}, w={})".format(
            self.x, self.y, self.z, self.w
        )

    def to_json(self):
        return {"x": self.x, "y": self.y, "z": self.z, "w": self.w}

    def from_json(self, json):
        self.x = json["x"]
        self.y = json["y"]
        self.z = json["z"]
        self.w = json["w"]


class Pose(Message):
    def __init__(self, position=Vector3(), orientation=Quaternion()):
        self.position = position  # type: Vector3
        self.orientation = orientation  # type: Quaternion

    def __str__(self):
        return "Pose(position={}, orientation={})".format(
            self.position, self.orientation
        )

    def to_json(self):
        return {
            "position": self.position.to_json(),
            "orientation": self.orientation.to_json(),
        }

    def from_json(self, json):
        self.position.from_json(json["position"])
        self.orientation.from_json(json["orientation"])


class PoseStamped(Message):
    def __init__(self, header=None, pose=Pose()):
        self.header = header  # type: Header
        self.pose = pose  # type: Pose

    def __str__(self):
        return "PoseStamped(header={}, pose={})".format(self.header, self.pose)

    def to_json(self):
        return {"header": self.header.to_json(), "pose": self.pose.to_json()}

    def from_json(self, json):
        self.header.from_json(json["header"])
        self.pose.from_json(json["pose"])


class PoseWithCovariance(Message):
    def __init__(self, pose=Pose(), covariance: Float64Array6x6 = []):
        self.pose = pose  # type: Pose
        self.covariance = covariance  # type: list

    def __str__(self):
        return "PoseWithCovariance(pose={}, covariance={})".format(
            self.pose, self.covariance
        )

    def to_json(self):
        return {"pose": self.pose.to_json(), "covariance": self.covariance}

    def from_json(self, json):
        self.pose.from_json(json["pose"])
        self.covariance = json["covariance"]


class TwistWithCovariance(Message):
    def __init__(self, twist=Twist(), covariance: Float64Array6x6 = []):
        self.twist = twist  # type: Twist
        self.covariance = covariance  # type: list

    def __str__(self):
        return "TwistWithCovariance(twist={}, covariance={})".format(
            self.twist, self.covariance
        )

    def to_json(self):
        return {"twist": self.twist.to_json(), "covariance": self.covariance}

    def from_json(self, json):
        self.twist.from_json(json["twist"])
        self.covariance = json["covariance"]



import threading

import zmq


class MessageBroker:
    def __init__(self):
        # gather list of topics from pub/subs
        self.topics = {}
        # run broker in seperate thread
        self.broker_thread = threading.Thread(target=self.config_broker)
        self.broker_thread.start()

    def config_broker(self):
        """connects, binds, and configure sockets"""
        # https://zguide.zeromq.org/docs/chapter2/#The-Dynamic-Discovery-Problem
        self.context = zmq.Context()
        frontend = self.context.socket(zmq.XSUB)
        frontend.bind("tcp://127.0.0.1:5559")
        backend = self.context.socket(zmq.XPUB)
        backend.bind("tcp://127.0.0.1:5560")

        # built-in pub/sub fowarder
        zmq.proxy(frontend, backend)

        # Shouldn't get here
        frontend.close()
        backend.close()
        self.context.term()


if __name__ == "__main__":
    MessageBroker()



import time

from . import Message


class Header:
    def __init__(self, seq: int = 0, stamp: float = None, frame_id: str = ""):
        self.seq = seq
        self.frame_id = frame_id
        self.stamp = stamp
        if stamp is None:
            self.stamp = time.time()

    def __str__(self):
        return "Header:\n  - seq={}\n  - stamp={}\n  - frame_id={})\n".format(
            self.seq, self.stamp, self.frame_id
        )

    def to_json(self):
        return {"seq": self.seq, "stamp": self.stamp, "frame_id": self.frame_id}

    def from_json(self, json):
        self.seq = json["seq"]
        self.stamp = json["stamp"]
        self.frame_id = json["frame_id"]


class GenericTypeData(Message):
    def __init__(self, data):
        self.data = data

    def __str__(self):
        return "String:\n  - data={}\n".format(self.data)

    def to_json(self):
        return {"data": self.data}

    def from_json(self, json):
        self.data = json["data"]


class String(GenericTypeData):
    def __init__(self, data: str = ""):
        super().__init__(data)


class Int(GenericTypeData):
    def __init__(self, data: int = 0):
        super().__init__(data)


class Float(GenericTypeData):
    def __init__(self, data: float = 0.0):
        super().__init__(data)


class Bool(GenericTypeData):
    def __init__(self, data: bool = False):
        super().__init__(data)



import asyncio
import json
import threading
import time

import zmq
import zmq.asyncio

from zeroros.messages import Header, Message
from zeroros.topic import validate_topic


class Subscriber:
    def __init__(
        self,
        topic: str,
        message_class: type[Message],
        callback_handle: callable,
        ip: str = "127.0.0.1",
        port: int = 5560,
    ):
        self.ip = ip
        self.port = port
        print("Subscribing to topic: ", topic)
        self.topic = validate_topic(topic)
        self.message_class = message_class
        self.url = "tcp://" + str(self.ip) + ":" + str(self.port)
        self.context = zmq.asyncio.Context()
        self.msg = self.message_class()
        self._stop = False
        self.listening_thread = threading.Thread(
            target=self.listen, args=(callback_handle,), daemon=True
        )
        self.listening_thread.start()

    def stop(self):
        self._stop = True

    def listen(self, callback_handle: callable):
        while not self._stop:
            try:
                self.msg = self.message_class()
                # create a task
                asyncio.run(self.recv_and_process())
                # register a done callback function
                callback_handle(self.msg)
            except Exception as e:
                print(f"Error for topic {self.topic} on {self.ip}:{self.port}: {e}")
            time.sleep(0.05)
        print("Stopping subscriber")
        # Check if the socket is still open
        if self.sock.closed is False:
            self.sock.close()
        self.context.term()

    async def recv_and_process(self):
        self.sock = self.context.socket(zmq.SUB)
        self.sock.setsockopt(zmq.SUBSCRIBE, str.encode(self.topic))
        self.sock.connect(self.url)
        received_msg = await self.sock.recv()
        self.sock.close()
        # Split the message into topic and message
        topic, message = received_msg.split(b" ", 1)
        # Check that the topic is correct
        if topic.decode("utf-8") != self.topic:
            raise ValueError(
                f"Topic received ({topic.decode('utf-8')})"
                f" does not match expected topic ({self.topic})"
            )
        try:
            # Decode the message
            message = message.decode("utf-8")
            # Parse the message
            message = json.loads(message)
            # Create a message object
            self.msg.from_json(message)
        except Exception as e:
            raise ValueError(f"Could not decode message: {e}")


if __name__ == "__main__":

    def callback(msg):
        print("Received in callback:", msg)

    sub = Subscriber("/header", Header, callback)
    while True:
        time.sleep(1)



from typing import List

from . import Message
from .geometry_msgs import PoseWithCovariance, TwistWithCovariance
from .std_msgs import Header


class Odometry(Message):
    def __init__(
        self,
        header=Header(),
        child_frame_id="",
        pose=PoseWithCovariance(),
        twist=TwistWithCovariance(),
    ):
        self.header = header  # type: Header
        self.child_frame_id = child_frame_id  # type: str
        self.pose = pose  # type: PoseWithCovariance
        self.twist = twist  # type: TwistWithCovariance

    def __str__(self):
        return "Odometry(header={}, child_frame_id={}, pose={}, twist={})".format(
            self.header, self.child_frame_id, self.pose, self.twist
        )

    def to_json(self):
        return {
            "header": self.header.to_json(),
            "child_frame_id": self.child_frame_id,
            "pose": self.pose.to_json(),
            "twist": self.twist.to_json(),
        }

    def from_json(self, json):
        self.header.from_json(json["header"])
        self.child_frame_id = json["child_frame_id"]
        self.pose.from_json(json["pose"])
        self.twist.from_json(json["twist"])


class Path(Message):
    def __init__(self, header=Header(), poses: List[PoseWithCovariance] = []):
        self.header = header  # type: Header
        self.poses = poses  # type: list

    def __str__(self):
        return "Path(header={}, poses={})".format(self.header, self.poses)

    def to_json(self):
        return {
            "header": self.header.to_json(),
            "poses": [pose.to_json() for pose in self.poses],
        }

    def from_json(self, json):
        self.header.from_json(json["header"])
        self.poses = [PoseWithCovariance().from_json(pose) for pose in json["poses"]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Copyright (c) 2023, Miquel Massot
All rights reserved.
Licensed under the GPLv3 License.
See LICENSE.md file in the project root for full license information.
"""

from pathlib import Path

from setuptools import find_packages, setup

GITHUB_URL = "https://github.com/miquelmassot/zeroros"


classifiers = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Developers",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: BSD License",
    "Programming Language :: Python :: 3",
    "Topic :: Software Development",
    "Topic :: Software Development :: Libraries",
    "Topic :: Software Development :: Libraries :: Python Modules",
]

# read the contents of README file
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="zeroros",
    version="1.0.0",
    description="ZeroROS middleware",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Miquel Massot",
    author_email="miquel.massot@gmail.com",
    maintainer="Miquel Massot",
    maintainer_email="miquel.massot@gmail.com",
    url=GITHUB_URL,
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    classifiers=classifiers,
    license="BSD-3-Clause",
    install_requires=[
        "numpy>=1.23.4",
        "pyzmq>=24.0.1",
        "setuptools>=59.6.0",
    ],
    project_urls={"Bug Reports": GITHUB_URL + "/issues", "Source": GITHUB_URL},
)
